from abc import abstractmethod
from pathlib import Path

from .base_sprite_loader import BaseSpriteLoader


class AbstractFloorSheet(BaseSpriteLoader):
    @property
    @abstractmethod
    def biome_name(self) -> str:
        raise NotImplementedError

    @property
    def _sprite_sheet_path(self) -> Path:
        return Path(f"Data/Textures/floor_{self.biome_name}.png")

    """
    Biome sheets are 12x12 128 pixel squares
    """
    _chunk_size = 128
    _chunk_map = {
        "push_block": (7, 0, 8, 1),
        "entrance": (0.5, 7, 3, 9),
        "entrance_shortcut": (0.5, 7, 3, 9),
        "door": (0.5, 7, 3, 9),
        # Exit is not on even 128 pixel boundaries RIP even numbers
        "exit": (0.5, 9.5, 3, 11.5),
        "door2": (8, 6, 10, 8),
        # door2_secret is the same as door2, but gets dirt/push_block in front
        "door2_secret": (8, 6, 10, 8),
        "spikes": (5, 9, 6, 10),
        "dirt": (0, 0, 4, 7),
        "floor": (0, 0, 1, 1),
        "ghist_door2": (10, 6, 12, 8),
    }
    _additional_chunks = {}

    def __init__(self, base_path: Path):
        super().__init__(base_path=base_path)
        # Adding additional chunks that are defined in subclasses to the chunk map
        new_map = self._chunk_map.copy()
        new_map.update(self._additional_chunks)
        self._chunk_map = new_map



from .base_classes.base_floorstyled_sheet import AbstractFloorStyledSheet

__all__ = [
    "HiveStyledFloorSheet",
    "DuatStyledFloorSheet",
    "GoldStyledFloorSheet",
    "GutsStyledFloorSheet",
    "PagodaStyledFloorSheet",
    "TempleStyledFloorSheet",
    "MothershipStyledFloorSheet",
    "BabylonStyledFloorSheet",
    "SunkenCityStyledFloorSheet",
    "PalaceStyledFloorSheet",
    "StonedStyledFloorSheet",
    "WoodStyledFloorSheet",
    "VladStyledFloorSheet",
]


class HiveStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "beehive"
    _additional_chunks = {
        "door2": (8, 6, 10, 8),
        "beehive_floor": (7, 2, 8, 3),
    }


class DuatStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "duat"
    _additional_chunks = {
        "duat_floor": (7, 2, 8, 3),
        "push_block": (7, 2, 8, 3),
    }


class GoldStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "gold"
    _additional_chunks = {
        "door2": (8, 6, 10, 8),
        "cog_floor": (7, 2, 8, 3),
        "slidingwall_ceiling": (5, 8, 6, 10),
        "slidingwall_switch": (6, 8, 7, 9),
        "crushtraplarge": (6, 0, 8, 2),
        "push_block": (7, 2, 8, 3),
    }


class GutsStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "guts"
    _additional_chunks = {
        "guts_floor": (7, 2, 8, 3),
    }


class MothershipStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "mothership"
    _additional_chunks = {
        "mothership_floor": (7, 2, 8, 3),
    }


class PagodaStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "pagoda"
    _additional_chunks = {
        "door2": (8, 6, 10, 8),
        "pagoda_floor": (7, 2, 8, 3),
        "slidingwall_ceiling": (5, 8, 6, 10),
        "slidingwall_switch": (6, 8, 7, 9),
    }


class BabylonStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "babylon"
    _additional_chunks = {
        "door2": (8, 6, 10, 8),
        "babylon_floor": (7, 2, 8, 3),
    }


class SunkenCityStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "sunken"
    _additional_chunks = {
        "door2": (8, 6, 10, 8),
        "sunken_floor": (7, 2, 8, 3),
    }


class PalaceStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "palace"
    _additional_chunks = {
        "palace_floor": (7, 2, 8, 3),
        "palace_entrance": (8, 6, 10, 8),
        "palace_table": (7, 1, 10, 2),
        "palace_table_tray": (8, 0, 9, 1),
        "palace_chandelier": (4, 8, 7, 10),
        "palace_candle": (7, 0, 8, 1),
        "palace_bookcase": (8, 2, 9, 3),
    }


class StonedStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "stone"
    _additional_chunks = {
        "door2": (8, 6, 10, 8),
        "stone_floor": (7, 2, 8, 3),
    }


class TempleStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "temple"
    _additional_chunks = {
        "door2": (8, 6, 10, 8),
        "temple_floor": (7, 2, 8, 3),
        "push_block": (7, 2, 8, 3),
    }


class VladStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "vlad"
    _additional_chunks = {
        "door2": (8, 6, 10, 8),
        "crown_statue": (8, 0, 10, 3),
        "vlad_floor": (7, 2, 8, 3),
    }


class WoodStyledFloorSheet(AbstractFloorStyledSheet):
    styled_name = "wood"
    _additional_chunks = {
        "door2": (8, 6, 10, 8),
        "door_drop_held": (8, 6, 10, 8),
        "minewood_floor": (7, 2, 8, 3),
        "wanted_poster": (6, 6, 8, 8),
        "shop_sign": (4, 8, 6, 10),
        "shop_door": (9, 0, 10, 1),
    }



import os
from queue import Empty

import tkinter as tk
from tkinter import ttk
from tkinter.scrolledtext import ScrolledText


# Adapted from https://beenje.github.io/blog/posts/logging-to-a-tkinter-scrolledtext-widget/
class ConsoleWindow(tk.Frame):
    def __init__(self, queue_handler, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Create a ScrolledText wdiget
        self.scrolled_text = ScrolledText(self, height=7, state="disabled")
        self.scrolled_text.pack(expand=True, fill="both")
        self.scrolled_text.configure(font="TkFixedFont")
        self.scrolled_text.tag_config("INFO", foreground="green")
        self.scrolled_text.tag_config("DEBUG", foreground="gray")
        self.scrolled_text.tag_config("WARNING", foreground="orange")
        self.scrolled_text.tag_config("ERROR", foreground="red")
        self.scrolled_text.tag_config("CRITICAL", foreground="red", underline=1)

        # Create a logging handler using a queue
        self.queue_handler = queue_handler

        # Start polling messages from the queue
        self.after(100, self.poll_log_queue)

    def display(self, record):
        msg = self.queue_handler.format(record)
        self.scrolled_text.configure(state="normal")
        self.scrolled_text.insert(tk.END, msg + "\n", record.levelname)
        self.scrolled_text.configure(state="disabled")
        self.scrolled_text.yview(tk.END)

    def poll_log_queue(self):
        # Check every 100ms if there is a new message in the queue to display
        while True:
            try:
                record = self.queue_handler.log_queue.get(block=False)
            except Empty:
                break
            else:
                self.display(record)
        self.after(100, self.poll_log_queue)

    def close(self):
        pass


class ToolTip:
    def __init__(self, widget, text):
        self.widget = widget
        self.text = text

        self.widget.bind("<Enter>", self.on_enter)
        self.widget.bind("<Leave>", self.on_leave)

        self.tooltip = None
        self.label = None

    def on_enter(self, event):
        self.tooltip = tk.Toplevel()
        self.tooltip.overrideredirect(True)
        self.tooltip.geometry(f"+{event.x_root+15}+{event.y_root+10}")

        self.label = tk.Label(self.tooltip, text=self.text)
        self.label.pack()

    def on_leave(self, _event):
        self.tooltip.destroy()


class Tab(ttk.Frame):
    """ Base class that all tabs should inherit from."""

    show_console = True

    def on_load(self):
        """ Called whenever the tab is loaded."""


class ScrollableFrame(ttk.LabelFrame):
    def __init__(self, container, *args, **kwargs):
        super().__init__(container, *args, **kwargs)
        self.canvas = tk.Canvas(self)
        self.scrollbar = ttk.Scrollbar(
            self, orient="vertical", command=self.canvas.yview
        )
        self.scrollable_frame = tk.Frame(self.canvas)
        self.canvas.pack(side="left", fill="both", expand=True)

        self.scrollable_frame.bind(
            "<Configure>",
            lambda e: self.canvas.configure(scrollregion=self.canvas.bbox("all")),
        )

        self.canvas.create_window((0, 0), window=self.scrollable_frame, anchor="nw")

        self.canvas.configure(yscrollcommand=self.scrollbar.set)
        self.canvas.bind("<Enter>", self._bind_to_mousewheel)
        self.canvas.bind("<Leave>", self._unbind_from_mousewheel)

        self.canvas.pack(side="left", fill="both", expand=True)
        self.scrollbar.pack(side="right", fill="y")

    def _on_mousewheel(self, event):
        scroll_dir = None
        if event.num == 5 or event.delta == -120:
            scroll_dir = 1
        elif event.num == 4 or event.delta == 120:
            scroll_dir = -1

        if scroll_dir is None:
            return

        # If the scrollbar is max size don't bother scrolling
        if self.scrollbar.get() == (0.0, 1.0):
            return

        self.canvas.yview_scroll(scroll_dir, "units")

    def _bind_to_mousewheel(self, _event):
        if "nt" in os.name:
            self.canvas.bind_all("<MouseWheel>", self._on_mousewheel)
        else:
            self.canvas.bind_all("<Button-4>", self._on_mousewheel)
            self.canvas.bind_all("<Button-5>", self._on_mousewheel)

    def _unbind_from_mousewheel(self, _event):
        if "nt" in os.name:
            self.canvas.unbind_all("<MouseWheel>")
        else:
            self.canvas.unbind_all("<Button-4>")
            self.canvas.unbind_all("<Button-5>")


class ToggledFrame(tk.Frame):
    def __init__(self, parent, text, *args, **options):
        tk.Frame.__init__(self, parent, *args, **options)

        self.show = tk.IntVar()
        self.show.set(0)

        self.title_frame = ttk.Frame(self)
        self.title_frame.pack(fill="x", expand=1)

        ttk.Label(self.title_frame, text=text).pack(side="left", fill="x", expand=1)

        self.toggle_button = ttk.Checkbutton(
            self.title_frame,
            width=2,
            text="+",
            command=self.toggle,
            variable=self.show,
            style="Toolbutton",
        )
        self.toggle_button.pack(side="left")

        self.sub_frame = tk.Frame(self, relief="sunken", borderwidth=1)

    def toggle(self):
        if bool(self.show.get()):
            self.sub_frame.pack(fill="x", expand=1)
            self.toggle_button.configure(text="-")
        else:
            self.sub_frame.forget()
            self.toggle_button.configure(text="+")



from pathlib import Path

from ..base_classes.base_sprite_merger import BaseSpriteMerger
from ..monsters.mounts import Mounts
from ..journal_mons import JournalMonsterSheet
from ..journal_items import JournalItemSheet
from ..items import ItemSheet
from ..util import chunks_from_animation


class TurkeySpriteMerger(BaseSpriteMerger):
    _target_sprite_sheet_path = Path("Data/Textures/Entities/turkey_full.png")
    _grid_hint_size = 8
    _origin_map = {
        Mounts: {
            **chunks_from_animation("turkey_row_0_", (0, 0, 1, 1), 14),
            **chunks_from_animation("turkey_row_1_", (0, 1, 1, 2), 16),
            **chunks_from_animation("turkey_row_2_", (0, 2, 1, 3), 16),
            **chunks_from_animation("turkey_row_3_", (0, 3, 1, 4), 14),
        },
        JournalMonsterSheet: {"journal_turkey": (0, 0, 1, 1)},
        JournalItemSheet: {"journal_cooked_turkey": (0, 0, 1, 1)},
        ItemSheet: {"cooked_turkey": (0, 0, 1, 1)},
    }


class RockdogSpriteMerger(BaseSpriteMerger):
    _target_sprite_sheet_path = Path("Data/Textures/Entities/rockdog_full.png")
    _grid_hint_size = 8
    _origin_map = {
        Mounts: {
            **chunks_from_animation("rockdog_row_0_", (0, 0, 1, 1), 14),
            **chunks_from_animation("rockdog_row_1_", (0, 1, 1, 2), 12),
            **chunks_from_animation("rockdog_row_2_", (0, 2, 1, 3), 15),
            **chunks_from_animation("rockdog_row_3_", (0, 3, 1, 4), 14),
        },
        JournalMonsterSheet: {"journal_rockdog": (0, 0, 1, 1)},
    }


class AxolotlSpriteMerger(BaseSpriteMerger):
    _target_sprite_sheet_path = Path("Data/Textures/Entities/axolotl_full.png")
    _grid_hint_size = 8
    _origin_map = {
        Mounts: {
            **chunks_from_animation("axolotl_row_0_", (0, 0, 1, 1), 14),
            **chunks_from_animation("axolotl_row_1_", (0, 1, 1, 2), 12),
            **chunks_from_animation("axolotl_row_2_", (0, 2, 1, 3), 15),
            **chunks_from_animation("axolotl_row_3_", (0, 3, 1, 4), 14),
        },
        JournalMonsterSheet: {"journal_axolotl": (0, 0, 1, 1)},
    }


class QilinSpriteMerger(BaseSpriteMerger):
    _target_sprite_sheet_path = Path("Data/Textures/Entities/qilin_full.png")
    _grid_hint_size = 8
    _origin_map = {
        Mounts: {
            **chunks_from_animation("qilin_row_0_", (0, 0, 1, 1), 14),
            **chunks_from_animation("qilin_row_1_", (0, 1, 1, 2), 12),
            **chunks_from_animation("qilin_row_2_", (0, 2, 1, 3), 15),
            **chunks_from_animation("qilin_row_3_", (0, 3, 1, 4), 14),
        },
        JournalMonsterSheet: {"journal_qilin": (0, 0, 1, 1)},
    }



import logging

logger = logging.getLogger("modlunky2")


class QueueHandler(logging.Handler):
    def __init__(self, log_queue):
        super().__init__()
        self.log_queue = log_queue

    def emit(self, record):
        self.log_queue.put(record)


def register_queue_handler(queue_handler):
    # On linux this handler exists in a subprocess but not on windows.
    for handler in logger.handlers:
        if isinstance(handler, QueueHandler):
            return

    formatter = logging.Formatter("%(asctime)s: %(message)s")
    queue_handler.setFormatter(formatter)
    logger.addHandler(queue_handler)
    logger.setLevel(logging.INFO)



import time
import logging
import tkinter as tk
from tkinter import PhotoImage, ttk
from multiprocessing import Queue

from modlunky2.constants import BASE_DIR, IS_EXE
from modlunky2.updater import self_update
from modlunky2.version import current_version, latest_version

from .tasks import TaskManager, PING_INTERVAL
from .config import ConfigTab
from .extract import ExtractTab
from .levels import LevelsTab
from .pack import PackTab
from .widgets import ConsoleWindow
from .logs import QueueHandler, register_queue_handler

logger = logging.getLogger("modlunky2")


MIN_WIDTH = 1280
MIN_HEIGHT = 900


class ModlunkyUI:
    def __init__(self, config, data_dir):
        self.config = config
        self.data_dir = data_dir

        self.current_version = current_version()
        self.latest_version = latest_version()

        if IS_EXE:
            if self.latest_version is None or self.current_version is None:
                self.needs_update = False
            else:
                self.needs_update = self.current_version < self.latest_version
        else:
            self.needs_update = False

        self._shutdown_handlers = []
        self._shutting_down = False

        self.log_queue = Queue()
        self.queue_handler = QueueHandler(self.log_queue)
        register_queue_handler(self.queue_handler)
        self.task_manager = TaskManager(self.log_queue)

        self.root = tk.Tk(className="Modlunky2")  # Equilux Black
        self.root.title("Modlunky 2")
        self.root.geometry(f"{MIN_WIDTH}x{MIN_HEIGHT}")
        self.root.minsize(MIN_WIDTH, MIN_HEIGHT)
        self.root.grid_columnconfigure(0, weight=1)
        self.root.grid_rowconfigure(0, weight=0)
        self.root.grid_rowconfigure(1, weight=1)
        self.root.grid_rowconfigure(2, weight=0)
        # self.root.resizable(False, False)
        self.icon_png = PhotoImage(file=BASE_DIR / "static/images/icon.png")
        self.root.iconphoto(False, self.icon_png)

        if self.needs_update:
            update_frame = ttk.LabelFrame(self.root, text="Modlunky2 Update Available")
            update_frame.grid(row=0, column=0, sticky="nswe")
            update_frame.columnconfigure(0, weight=1)
            update_frame.rowconfigure(0, weight=1)

            update_button = tk.Button(
                update_frame,
                text="Update Now!",
                command=self.update,
                bg="#bfbfbf",
                font="sans 12 bold",
            )
            update_button.grid(column=0, row=0, sticky="nswe")

        # Handle shutting down cleanly
        self.root.protocol("WM_DELETE_WINDOW", self.quit)

        self.tabs = {}
        self.tab_control = ttk.Notebook(self.root)

        self.register_tab(
            "Pack Assets",
            PackTab(
                tab_control=self.tab_control,
                config=config,
                task_manager=self.task_manager,
            ),
        )
        self.register_tab(
            "Extract Assets",
            ExtractTab(
                tab_control=self.tab_control,
                config=config,
                task_manager=self.task_manager,
            ),
        )
        self.register_tab(
            "Level Editor",
            LevelsTab(
                tab_control=self.tab_control,
                config=config,
            ),
        )

        self.register_tab(
            "Config",
            ConfigTab(
                tab_control=self.tab_control,
                config=config,
            ),
        )

        self.tab_control.bind("<<NotebookTabChanged>>", self.on_tab_change)
        self.tab_control.grid(column=0, row=1, padx=2, pady=(4, 0), sticky="nsew")

        self.console_frame = ttk.LabelFrame(self.root, text="Console")
        self.console_frame.columnconfigure(0, weight=1)
        self.console_frame.rowconfigure(0, weight=1)

        self.console = ConsoleWindow(self.queue_handler, self.console_frame)
        self.console.grid(column=0, row=0, padx=2, pady=2, sticky="ew")

        self.version_label = tk.Label(
            self.root, text=f"v{self.current_version}", font="Helvitica 9 italic"
        )
        self.version_label.grid(column=0, row=3, padx=5, sticky="e")

        self.task_manager.start_process()
        self.last_ping = time.time()
        self.root.after(100, self.after_cb)

    def after_cb(self):
        if not self.task_manager.is_alive():
            # Worker process went away but we're shutting down so just return
            if self._shutting_down:
                return

            # Worker process went away unexpectedly... Restart it.
            logger.critical("Worker process went away... Restarting it.")
            self.task_manager.start_process()
            self.root.after(100, self.after_cb)
            return

        # Send regular pings so the worker process knows
        # we're still alive.
        now = time.time()
        if now - self.last_ping > PING_INTERVAL:
            self.last_ping = now
            self.task_manager.ping()

        while True:
            msg = self.task_manager.receive_message()
            if msg is None:
                self.root.after(100, self.after_cb)
                return

            self.task_manager.dispatch(msg)

    def update(self):
        try:
            self_update()
            self.quit()
        except Exception:  # pylint: disable=broad-except
            logger.exception("Failed to update...")

    def on_tab_change(self, event):
        tab_name = event.widget.tab("current")["text"]
        tab = self.tabs[tab_name]
        if tab.show_console:
            self.console_frame.grid(row=2, column=0, padx=5, pady=(5, 0), sticky="nswe")
        else:
            self.console_frame.grid_forget()
        tab.on_load()

    def register_tab(self, name, obj):
        self.tabs[name] = obj
        self.tab_control.add(obj, text=name)

    def quit(self):
        if self._shutting_down:
            return

        self._shutting_down = True
        logger.info("Shutting Down.")
        for handler in self._shutdown_handlers:
            handler()

        self.task_manager.quit()
        self.root.quit()
        self.root.destroy()

    def register_shutdown_handler(self, func):
        self._shutdown_handlers.append(func)

    def mainloop(self):
        try:
            self.root.mainloop()
        except KeyboardInterrupt:
            self.quit()



import logging
import shutil
import tkinter as tk
from pathlib import Path
from tkinter import ttk
from ctypes.util import find_library  # pylint: disable=unused-import

from fsb5.utils import load_lib, LibraryNotFoundException

from modlunky2.assets.assets import AssetStore
from modlunky2.assets.constants import (
    EXTRACTED_DIR,
    FILEPATH_DIRS,
    OVERRIDES_DIR,
    PACKS_DIR,
)
from modlunky2.ui.utils import is_patched
from modlunky2.ui.widgets import Tab, ToolTip
from modlunky2.assets.soundbank import Extension as SoundExtension

logger = logging.getLogger("modlunky2")


MODS = Path("Mods")

TOP_LEVEL_DIRS = [EXTRACTED_DIR, PACKS_DIR, OVERRIDES_DIR]


def try_load_vorbis():
    try:
        load_lib("vorbis")
        return True
    except LibraryNotFoundException:
        return False


def extract_assets(
    _call,
    install_dir,
    target,
    recompress,
    generate_string_hashes,
    create_entity_sheets,
    extract_sound_extensions,
    reuse_extracted,
):
    exe_filename = install_dir / target

    if is_patched(exe_filename):
        logger.critical(
            (
                "%s is a patched exe. Can't extract. You should Restore Exe"
                " or validate game files to get a clean exe before Extracting."
            ),
            exe_filename,
        )
        return

    mods_dir = install_dir / MODS

    for dir_ in TOP_LEVEL_DIRS:
        (mods_dir / dir_).mkdir(parents=True, exist_ok=True)

    for dir_ in FILEPATH_DIRS:
        (mods_dir / EXTRACTED_DIR / dir_).mkdir(parents=True, exist_ok=True)
        (mods_dir / ".compressed" / EXTRACTED_DIR / dir_).mkdir(
            parents=True, exist_ok=True
        )

    with exe_filename.open("rb") as exe:
        asset_store = AssetStore.load_from_file(exe)
        unextracted = asset_store.extract(
            mods_dir / EXTRACTED_DIR,
            mods_dir / ".compressed" / EXTRACTED_DIR,
            recompress=recompress,
            generate_string_hashes=generate_string_hashes,
            create_entity_sheets=create_entity_sheets,
            extract_sound_extensions=extract_sound_extensions,
            reuse_extracted=reuse_extracted,
        )

    for asset in unextracted:
        logger.warning("Un-extracted Asset %s", asset.asset_block)

    dest = mods_dir / EXTRACTED_DIR / "Spel2.exe"
    if not reuse_extracted and exe_filename != dest:
        logger.info("Backing up exe to %s", dest)
        shutil.copy2(exe_filename, dest)

    logger.info("Extraction complete!")


class ExtractTab(Tab):
    def __init__(self, tab_control, config, task_manager, *args, **kwargs):
        super().__init__(tab_control, *args, **kwargs)
        self.tab_control = tab_control
        self.config = config
        self.task_manager = task_manager
        self.task_manager.register_task(
            "extract_assets",
            extract_assets,
            True,
            on_complete="extract_finished",
        )
        self.task_manager.register_handler("extract_finished", self.extract_finished)

        self.vorbis_loaded = try_load_vorbis()

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, minsize=60)

        self.top_frame = ttk.Frame(self)
        self.top_frame.rowconfigure(0, weight=1)
        self.top_frame.columnconfigure(0, weight=1)
        self.top_frame.columnconfigure(1, minsize=250)
        self.top_frame.grid(row=0, column=0, sticky="nswe")

        self.exe_frame = ttk.LabelFrame(self.top_frame, text="Select exe to Extract")
        self.exe_frame.grid(row=0, column=0, pady=5, padx=5, sticky="nswe")
        self.exe_frame.rowconfigure(0, weight=1)
        self.exe_frame.columnconfigure(0, weight=1)
        self.exe_frame.columnconfigure(1)

        self.list_box = tk.Listbox(self.exe_frame)
        self.list_box.grid(row=0, column=0, sticky="nswe")
        self.scrollbar = ttk.Scrollbar(self.exe_frame)
        self.scrollbar.grid(row=0, column=1, sticky="nes")

        self.config_frame = ttk.LabelFrame(self.top_frame, text="Options")
        self.config_frame.grid(row=0, column=1, pady=5, padx=5, sticky="nswe")

        self.recompress = tk.BooleanVar()
        self.recompress.set(True)
        self.checkbox_recompress = tk.Checkbutton(
            self.config_frame,
            text="Recompress",
            variable=self.recompress,
            onvalue=True,
            offvalue=False,
        )
        self.checkbox_recompress.grid(row=0, sticky="nw")
        ToolTip(
            self.checkbox_recompress,
            (
                "Recompress assets to speed up futuring packing.\n"
                "Not necessary if you just want the extracted assets."
            ),
        )

        self.create_entity = tk.BooleanVar()
        self.create_entity.set(True)
        self.checkbox_create_entity = tk.Checkbutton(
            self.config_frame,
            text="Create Entity Sprites",
            variable=self.create_entity,
            onvalue=True,
            offvalue=False,
        )
        self.checkbox_create_entity.grid(row=1, sticky="nw")
        ToolTip(
            self.checkbox_create_entity,
            (
                "Create merged entity spritesheets. These provide a simpler\n"
                "interface to some entity mods."
            ),
        )

        self.generate_string_hashes = tk.BooleanVar()
        self.generate_string_hashes.set(True)
        self.checkbox_string_hashes = tk.Checkbutton(
            self.config_frame,
            text="Generate String Hashes",
            variable=self.generate_string_hashes,
            onvalue=True,
            offvalue=False,
        )
        self.checkbox_string_hashes.grid(row=2, sticky="nw")
        ToolTip(
            self.checkbox_string_hashes,
            (
                "Generate string hash files. These provide a better\n"
                "interface for creating mods that use strings."
            ),
        )

        self.extract_wavs = tk.BooleanVar()
        self.extract_wavs.set(False)
        self.checkbox_extract_wavs = tk.Checkbutton(
            self.config_frame,
            text="Extract .wav files",
            variable=self.extract_wavs,
            onvalue=True,
            offvalue=False,
        )
        self.checkbox_extract_wavs.grid(row=3, sticky="nw")
        ToolTip(self.checkbox_extract_wavs, ("Extract .wav files from the soundbank."))

        oggs_state = tk.NORMAL
        if not self.vorbis_loaded:
            oggs_state = tk.DISABLED
        self.extract_oggs = tk.BooleanVar()
        self.extract_oggs.set(False)
        self.checkbox_extract_oggs = tk.Checkbutton(
            self.config_frame,
            state=oggs_state,
            text="Extract .ogg files",
            variable=self.extract_oggs,
            onvalue=True,
            offvalue=False,
        )
        self.checkbox_extract_oggs.grid(row=4, sticky="nw")
        ToolTip(self.checkbox_extract_oggs, ("Extract .ogg files from the soundbank."))

        self.reuse_extracted = tk.BooleanVar()
        self.reuse_extracted.set(False)
        self.checkbox_reuse_extracted = tk.Checkbutton(
            self.config_frame,
            text="Reuse Extracted Assets",
            variable=self.reuse_extracted,
            onvalue=True,
            offvalue=False,
        )
        self.checkbox_reuse_extracted.grid(row=5, sticky="nw")
        ToolTip(
            self.checkbox_reuse_extracted,
            (
                "If checked we will not re-extract from the binary\n"
                "and only rerun post processing steps."
            ),
        )

        self.list_box.config(yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.list_box.yview)

        self.button_extract = ttk.Button(self, text="Extract", command=self.extract)
        self.button_extract.grid(row=1, column=0, pady=5, padx=5, sticky="nswe")
        ToolTip(self.button_extract, ("Extract assets from EXE."))

    def extract_finished(self):
        self.button_extract["state"] = tk.NORMAL

    def extract(self):
        idx = self.list_box.curselection()
        if not idx:
            logger.error("Didn't select exe")
            return

        self.button_extract["state"] = tk.DISABLED

        extract_sound_extensions = []
        if self.extract_wavs.get():
            extract_sound_extensions.append(SoundExtension.WAV)

        if self.extract_oggs.get():
            extract_sound_extensions.append(SoundExtension.OGG)

        selected_exe = self.list_box.get(idx)
        self.task_manager.call(
            "extract_assets",
            install_dir=self.config.install_dir,
            target=selected_exe,
            recompress=self.recompress.get(),
            generate_string_hashes=self.generate_string_hashes.get(),
            create_entity_sheets=self.create_entity.get(),
            extract_sound_extensions=extract_sound_extensions,
            reuse_extracted=self.reuse_extracted.get(),
        )

    def get_exes(self):
        exes = []
        # Don't recurse forever. 3 levels should be enough
        exes.extend(self.config.install_dir.glob("*.exe"))
        exes.extend(self.config.install_dir.glob("*/*.exe"))
        exes.extend(self.config.install_dir.glob("*/*/*.exe"))
        return [
            exe.relative_to(self.config.install_dir)
            for exe in exes
            # Exclude modlunky2 which is likely in the install directory
            if exe.name not in ["modlunky2.exe"]
        ]

    def on_load(self):
        self.list_box.delete(0, tk.END)
        for exe in self.get_exes():
            self.list_box.insert(tk.END, str(exe))



from .base_classes import AbstractFloorSheet

__all__ = [
    "CaveFloorSheet",
    "VolcanaFloorSheet",
    "JungleFloorSheet",
    "TidePoolFloorSheet",
    "TempleFloorSheet",
    "IceCavesFloorSheet",
    "SurfaceFloorSheet",
    "SunkenCityFloorSheet",
    "BabylonFloorSheet",
    "EggplantFloorSheet",
]


class CaveFloorSheet(AbstractFloorSheet):
    biome_name = "cave"
    _additional_chunks = {
        "bush_block": (10, 2, 11, 3),
        "ladder": (4, 1, 5, 2),
        "ladder_plat": (4, 2, 5, 3),
        "bone_block": (10, 2, 11, 3),
        "idol_floor": (10, 0, 11, 1),
        "starting_exit": (0.5, 7, 3, 9),
        "minewood_floor_hanging_hide": (6, 1, 7, 3),
    }


class VolcanaFloorSheet(AbstractFloorSheet):
    biome_name = "volcano"
    _additional_chunks = {
        "icefloor": (7, 1, 8, 2),
        "conveyorbelt_right": (8, 10, 9, 11),
        "conveyorbelt_left": (8, 11, 9, 12),
        "door2": (8, 6, 10, 8),
        "falling_platform": (4, 5, 5, 6),
    }


class JungleFloorSheet(AbstractFloorSheet):
    biome_name = "jungle"
    _additional_chunks = {
        "vine": (4, 1, 5, 2),
        "growable_vine": (4, 2, 5, 3),
        "tree_base": (3, 11, 4, 12),
        "jungle_floor": (0, 0, 1, 1),
    }


class TidePoolFloorSheet(AbstractFloorSheet):
    biome_name = "tidepool"
    _additional_chunks = {
        "climbing_pole": (4, 1, 5, 2),
        "growable_climbing_pole": (4, 0, 5, 1),
        "fountain_head": (10, 2, 12, 4),
    }


class TempleFloorSheet(AbstractFloorSheet):
    biome_name = "temple"


class IceCavesFloorSheet(AbstractFloorSheet):
    biome_name = "ice"
    _additional_chunks = {
        "icefloor": (7, 1, 8, 2),
        "thinice": (3, 11, 4, 12),
        "upsidedown_spikes": (8, 9, 9, 10),
        "falling_platform": (4, 5, 5, 6),
        "eggplant_altar": (10, 2, 11, 3),
    }


class BabylonFloorSheet(AbstractFloorSheet):
    biome_name = "babylon"
    _additional_chunks = {
        "mushroom_base": (9, 11, 10, 12),
    }


class SunkenCityFloorSheet(AbstractFloorSheet):
    biome_name = "sunken"
    _additional_chunks = {"bigspear_trap": (8, 9, 10, 10)}


class EggplantFloorSheet(AbstractFloorSheet):
    biome_name = "eggplant"
    _additional_chunks = {
        "eggplant_door": (0.5, 7, 3, 9),
    }


class SurfaceFloorSheet(AbstractFloorSheet):
    biome_name = "surface"
    _additional_chunks = {"surface_floor": (0, 0, 1, 1)}



from pathlib import Path

from .base_classes import BaseSpriteLoader


class JournalItemSheet(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/journal_entry_items.png")
    _chunk_size = 160
    _chunk_map = {
        "journal_rope": (0, 0, 1, 1),
        "journal_bomb_bag": (1, 0, 2, 1),
        "journal_bomb_box": (2, 0, 3, 1),
        "journal_paste": (3, 0, 4, 1),
        "journal_speckles": (4, 0, 5, 1),
        "journal_climbing_gloves": (5, 0, 6, 1),
        "journal_pitchers_mitt": (6, 0, 7, 1),
        "journal_spring_boots": (7, 0, 8, 1),
        "journal_spike_boots": (8, 0, 9, 1),
        "journal_compass": (9, 0, 10, 1),
        "journal_parachute": (0, 1, 1, 2),
        "journal_udjat_eye": (1, 1, 2, 2),
        "journal_kapala": (2, 1, 3, 2),
        "journal_hedjet": (3, 1, 4, 2),
        "journal_crown": (4, 1, 5, 2),
        "journal_ankh": (5, 1, 6, 2),
        "journal_royal_jelly": (6, 1, 7, 2),
        "journal_cape": (7, 1, 8, 2),
        "journal_jetpack": (8, 1, 9, 2),
        "journal_telepack": (9, 1, 10, 2),
        "journal_webgun": (0, 2, 1, 3),
        "journal_shotgun": (1, 2, 2, 3),
        "journal_freeze_ray": (2, 2, 3, 3),
        "journal_crossbow": (3, 2, 4, 3),
        "journal_camera": (4, 2, 5, 3),
        "journal_teleporter": (5, 2, 6, 3),
        "journal_mattock": (6, 2, 7, 3),
        "journal_boomerang": (7, 2, 8, 3),
        "journal_machete": (8, 2, 9, 3),
        "journal_excalibur": (9, 2, 10, 3),
        "journal_plasma_canon": (0, 3, 1, 4),
        "journal_scepter": (1, 3, 2, 4),
        "journal_idol": (2, 3, 3, 4),
        "journal_ushtabi": (3, 3, 4, 4),
        "journal_eggplant": (4, 3, 5, 4),
        "journal_cooked_turkey": (5, 3, 6, 4),
        "journal_hover_pack": (6, 3, 7, 4),
        "journal_clone_gun": (7, 3, 8, 4),
        "journal_skeleton_key": (8, 3, 9, 4),
        "journal_tablet_of_destiny": (9, 3, 10, 4),
        "journal_wooden_shield": (0, 4, 1, 5),
        "journal_metal_shield": (1, 4, 2, 5),
        "journal_clover": (2, 4, 3, 5),
        "journal_the_true_crown": (3, 4, 4, 5),
        "journal_power_pack": (4, 4, 5, 5),
        "journal_vlads_cape": (5, 4, 6, 5),
        "journal_elixir": (6, 4, 7, 5),
        "journal_tusk_idol": (7, 4, 8, 5),
        "journal_alien_compass": (8, 4, 9, 5),
        "journal_curse_pot": (9, 4, 10, 5),
        "journal_hou_yis_bow": (0, 5, 1, 6),
        "journal_arrow_of_light": (1, 5, 2, 6),
        "journal_broken_sword": (2, 5, 3, 6),
        "journal_eggplant_crown": (3, 5, 4, 6),
        "journal_blood_splat_1": (4, 5, 5, 6),
        "journal_blood_splat_2": (5, 5, 6, 6),
        "journal_blood_splat_3": (6, 5, 7, 6),
        "journal_dog_paw_print": (7, 5, 8, 6),
        "journal_cat_paw_print": (8, 5, 9, 6),
        "journal_hamster_paw_print": (9, 5, 10, 6),
        # TODO: Bottom part
    }



from .character import (
    CharacterBlackSpriteMerger,
    CharacterLimeSpriteMerger,
    CharacterMagentaSpriteMerger,
    CharacterOliveSpriteMerger,
    CharacterOrangeSpriteMerger,
    CharacterPinkSpriteMerger,
    CharacterRedSpriteMerger,
    CharacterVioletSpriteMerger,
    CharacterWhiteSpriteMerger,
    CharacterYellowSpriteMerger,
    CharacterBlueSpriteMerger,
    CharacterCeruleanSpriteMerger,
    CharacterCinnabarSpriteMerger,
    CharacterCyanSpriteMerger,
    CharacterGoldSpriteMerger,
    CharacterGraySpriteMerger,
    CharacterGreenSpriteMerger,
    CharacterIrisSpriteMerger,
    CharacterKhakiSpriteMerger,
    CharacterLemonSpriteMerger,
    CharacterEggChildSpriteMerger,
    CharacterHiredHandSpriteMerger,
)
from .mounts import (
    TurkeySpriteMerger,
    RockdogSpriteMerger,
    AxolotlSpriteMerger,
    QilinSpriteMerger,
)
from .pets import MontySpriteMerger, PercySpriteMerger, PoochiSpriteMerger

__all__ = [
    "CharacterBlackSpriteMerger",
    "CharacterLimeSpriteMerger",
    "CharacterMagentaSpriteMerger",
    "CharacterOliveSpriteMerger",
    "CharacterOrangeSpriteMerger",
    "CharacterPinkSpriteMerger",
    "CharacterRedSpriteMerger",
    "CharacterVioletSpriteMerger",
    "CharacterWhiteSpriteMerger",
    "CharacterYellowSpriteMerger",
    "CharacterBlueSpriteMerger",
    "CharacterCeruleanSpriteMerger",
    "CharacterCinnabarSpriteMerger",
    "CharacterCyanSpriteMerger",
    "CharacterGoldSpriteMerger",
    "CharacterGraySpriteMerger",
    "CharacterGreenSpriteMerger",
    "CharacterIrisSpriteMerger",
    "CharacterKhakiSpriteMerger",
    "CharacterLemonSpriteMerger",
    "CharacterEggChildSpriteMerger",
    "CharacterHiredHandSpriteMerger",
    "TurkeySpriteMerger",
    "RockdogSpriteMerger",
    "AxolotlSpriteMerger",
    "QilinSpriteMerger",
    "MontySpriteMerger",
    "PercySpriteMerger",
    "PoochiSpriteMerger",
]



import logging
import json

try:
    import winreg
except ImportError:
    winreg = None

from shutil import copyfile
from pathlib import Path

from modlunky2.constants import APP_DIR


PROGRAMS_KEY = "Software\\Microsoft\\Windows\\CurrentVersion\\Uninstall"
DEFAULT_PATH = Path("C:/Program Files (x86)/Steam/steamapps/common/Spelunky 2")
EXE_NAME = "Spel2.exe"

# Sentinel for tracking unset fields
NOT_PRESENT = object()

logger = logging.getLogger("modlunky2")


def check_registry_for_spel2():
    if winreg is None:
        return None

    programs = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, PROGRAMS_KEY)

    index = 0
    while True:
        try:
            keyname = winreg.EnumKey(programs, index)
            index += 1
        except OSError:
            return None

        try:
            subkey = winreg.OpenKey(programs, keyname)
            name, _ = winreg.QueryValueEx(subkey, "DisplayName")
            if name == "Spelunky 2":
                return Path(winreg.QueryValueEx(subkey, "InstallLocation")[0])
        except OSError:
            continue

    return None


def guess_install_dir():
    logger.info("Checking if Spelunky 2 is installed in %s", APP_DIR)
    if (APP_DIR / EXE_NAME).exists():
        logger.info("Found Spelunky 2!")
        return APP_DIR

    logger.info("Checking if Spelunky 2 is installed in %s", DEFAULT_PATH)
    if (DEFAULT_PATH / EXE_NAME).exists():
        logger.info("Found Spelunky 2!")
        return DEFAULT_PATH

    logger.info("Checking Windows Registry for Spelunky 2 Installation")
    reg_path = check_registry_for_spel2()
    if reg_path and (reg_path / EXE_NAME).exists():
        logger.info("Found Spelunky 2!")
        return reg_path

    logger.warning("No Spelunky 2 installation found...")
    return None


class ConfigFile:
    def __init__(self, config_path: Path):
        self.config_path = config_path

        self.install_dir = None

    @classmethod
    def from_path(cls, config_path: Path):
        obj = cls(config_path=config_path)
        needs_save = False

        if config_path.exists():
            with config_path.open("r") as config_file:
                config_data = json.load(config_file)
        else:
            config_data = {}

        # Initialize install-dir
        install_dir = config_data.get("install-dir", NOT_PRESENT)
        if install_dir is NOT_PRESENT:
            install_dir = guess_install_dir()
            needs_save = True
        elif install_dir is not None:
            install_dir = Path(install_dir)
        obj.install_dir = install_dir

        if needs_save:
            obj.save()

        return obj

    def to_dict(self):
        install_dir = None
        if self.install_dir:
            install_dir = self.install_dir.as_posix()

        return {
            "install-dir": install_dir,
        }

    def _get_tmp_path(self):
        return self.config_path.with_suffix(f"{self.config_path.suffix}.tmp")

    def save(self):
        # Make a temporary file so we can do an atomic replace
        # in case something crashes while writing out the config.
        tmp_path = self._get_tmp_path()
        with tmp_path.open("w") as tmp_file:
            json.dump(self.to_dict(), tmp_file)

        copyfile(tmp_path, self.config_path)
        tmp_path.unlink()


class Config:
    CONFIG_FIELDS = set(["install_dir"])

    def __init__(self, config_file: ConfigFile):
        self.config_file = config_file

        self._install_dir = NOT_PRESENT
        self.beta = False

    @classmethod
    def from_path(cls, config_path: Path):
        return cls(config_file=ConfigFile.from_path(config_path))

    @property
    def install_dir(self):
        if self._install_dir is NOT_PRESENT:
            if self.config_file.install_dir is None:
                return APP_DIR
            return self.config_file.install_dir
        return self._install_dir

    @install_dir.setter
    def install_dir(self, value):
        self._install_dir = value



import logging
from pathlib import Path

import tkinter as tk
from tkinter import ttk

from modlunky2.ui.widgets import Tab, ToolTip
from modlunky2.config import guess_install_dir

logger = logging.getLogger("modlunky2")


class ConfigTab(Tab):
    def __init__(self, tab_control, config, *args, **kwargs):
        super().__init__(tab_control, *args, **kwargs)
        self.tab_control = tab_control
        self.config = config

        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        config_frame = ttk.LabelFrame(self, text="Config")
        config_frame.rowconfigure(0, weight=1)
        config_frame.columnconfigure(0, weight=1)
        config_frame.grid(row=0, column=0, pady=5, padx=5, sticky="nswe")

        install_dir_frame = ttk.LabelFrame(config_frame, text="Install Directory")
        install_dir_frame.grid(row=0, column=0, pady=5, padx=5, sticky="n")
        ToolTip(install_dir_frame, ("The directory where Spelunky 2 is installed."))

        self.install_dir_var = tk.StringVar(value=self.config.config_file.install_dir)
        install_dir_entry = ttk.Entry(
            install_dir_frame,
            textvariable=self.install_dir_var,
            state=tk.DISABLED,
            width=80,
        )
        install_dir_entry.columnconfigure(0, weight=1)
        install_dir_entry.columnconfigure(1, weight=1)
        install_dir_entry.columnconfigure(2, weight=1)
        install_dir_entry.grid(
            row=0, column=0, padx=10, pady=10, columnspan=3, sticky="n"
        )

        install_dir_browse = ttk.Button(
            install_dir_frame, text="Browse", command=self.browse_install_dir
        )
        install_dir_browse.grid(row=1, column=0, pady=5, padx=5, sticky="nsew")

        install_dir_lucky = ttk.Button(
            install_dir_frame, text="I'm Feeling Lucky", command=self.feeling_lucky
        )
        install_dir_lucky.grid(row=1, column=1, pady=5, padx=5, sticky="nsew")

        install_dir_clear = ttk.Button(
            install_dir_frame, text="Clear", command=self.clear_install_dir
        )
        install_dir_clear.grid(row=1, column=2, pady=5, padx=5, sticky="nsew")

    def browse_install_dir(self):
        install_dir = self.install_dir_var.get()
        directory = tk.filedialog.askdirectory(initialdir=install_dir)
        if directory:
            self.install_dir_var.set(directory)
            self.config.config_file.install_dir = Path(directory)
            self.config.config_file.save()

    def feeling_lucky(self):
        install_dir = guess_install_dir()
        if install_dir:
            self.install_dir_var.set(install_dir)
            self.config.config_file.install_dir = install_dir
            self.config.config_file.save()

    def clear_install_dir(self):
        self.install_dir_var.set("")
        self.config.config_file.install_dir = None
        self.config.config_file.save()



from pathlib import Path

from .base_classes import BaseSpriteLoader


class EggShip2Sheet(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/base_eggship2.png")
    _chunk_size = 128
    _chunk_map = {
        "olmecship": (0, 0, 3, 3),
    }



from pathlib import Path

from .base_classes import BaseSpriteLoader
from .util import chunks_from_animation


class CharacterSheet(BaseSpriteLoader):
    _chunk_size = 128
    _chunk_map = {
        "portrait": (12, 1, 16, 7),
        "standing": (0, 0, 1, 1),
        **chunks_from_animation("walking", (1, 0, 2, 1), 8),
        "stunned": (9, 0, 10, 1),
        **chunks_from_animation("swimming", (10, 0, 11, 1), 6),
        **chunks_from_animation("crouching", (0, 1, 1, 2), 5),
        **chunks_from_animation("crawling", (5, 1, 6, 2), 7),
        **chunks_from_animation("stun", (0, 2, 1, 3), 4),
        **chunks_from_animation("flip", (4, 2, 5, 3), 7),
        "holding": (11, 2, 12, 3),
        **chunks_from_animation("teetering", (0, 3, 1, 4), 8),
        **chunks_from_animation("ledge_grab", (8, 3, 9, 4), 4),
        **chunks_from_animation("whipping", (0, 4, 1, 5), 6),
        **chunks_from_animation("throwing", (6, 4, 7, 5), 5),
        **chunks_from_animation("enter", (0, 5, 1, 6), 6),
        **chunks_from_animation("exit", (6, 5, 7, 6), 6),
        **chunks_from_animation("climbing_ladder", (0, 6, 1, 7), 6),
        **chunks_from_animation("pushing", (6, 6, 7, 7), 6),
        **chunks_from_animation("climbing_rope", (0, 7, 1, 8), 10),
        **chunks_from_animation("crouching_mount", (10, 7, 11, 8), 5),
        **chunks_from_animation("looking_up", (0, 8, 1, 9), 7),
        "sitting": (7, 8, 8, 9),
        **chunks_from_animation("looking_up_sitting", (8, 8, 9, 9), 7),
        **chunks_from_animation("jumping", (0, 9, 1, 10), 12),
        "rope_coiled": (12, 9, 13, 10),
        "rope_top": (13, 9, 14, 10),
        "rope_short": (14, 9, 15, 10),
        "rope_burned": (15, 9, 16, 10),
        **chunks_from_animation("ghosting", (0, 10, 1, 11), 16),
        "icon": (0, 11, 1, 12),
        "bag": (1, 11, 2, 12),
        "wanted": (2, 11, 3, 12),
        "pointer": (3, 11, 4, 12),
        **chunks_from_animation("taming", (4, 11, 5, 12), 6),
        "ghost_stun": (10, 11, 11, 12),
        "rope_segment": (0, 12, 1, 13),
        **chunks_from_animation("rope_uncoiling", (1, 12, 2, 13), 4),
        "rope_end": (5, 12, 6, 13),
        **chunks_from_animation("rope_burning", (6, 12, 7, 13), 4),
        **chunks_from_animation("whip", (10, 12, 11, 13), 6),
        **chunks_from_animation("stun_effect", (0, 13, 1, 14), 12),
        **chunks_from_animation("petting", (0, 14, 1, 15), 8),
    }


class CharacterBlackSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_black.png")


class CharacterLimeSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_lime.png")


class CharacterMagentaSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_magenta.png")


class CharacterOliveSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_olive.png")


class CharacterOrangeSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_orange.png")


class CharacterPinkSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_pink.png")


class CharacterRedSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_red.png")


class CharacterVioletSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_violet.png")


class CharacterWhiteSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_white.png")


class CharacterYellowSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_yellow.png")


class CharacterBlueSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_blue.png")


class CharacterCeruleanSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_cerulean.png")


class CharacterCinnabarSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_cinnabar.png")


class CharacterCyanSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_cyan.png")


class CharacterEggChildSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_eggchild.png")


class CharacterGoldSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_gold.png")


class CharacterGraySheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_gray.png")


class CharacterGreenSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_green.png")


class CharacterHiredHandSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_hired.png")


class CharacterIrisSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_iris.png")


class CharacterKhakiSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_khaki.png")


class CharacterLemonSheet(CharacterSheet):
    _sprite_sheet_path = Path("Data/Textures/char_lemon.png")



import re
from dataclasses import dataclass
from collections import OrderedDict
from typing import ClassVar, Optional, TextIO

from .utils import DirectivePrefixes, split_comment, to_line

VALID_TILE_CODES = set(
    [
        "alien",
        "adjacent_floor",
        "alien_generator",
        "alienqueen",
        "altar",
        "ammit",
        "ankh",
        "anubis",
        "arrow_trap",
        "autowalltorch",
        "babylon_floor",
        "beehive_floor",
        "bigspear_trap",
        "bodyguard",
        "bone_block",
        "bunkbed",
        "bush_block",
        "catmummy",
        "caveman",
        "caveman_asleep",
        "cavemanboss",
        "cavemanshopkeeper",
        "chain_ceiling",
        "chainandblocks_ceiling",
        "chair_looking_left",
        "chair_looking_right",
        "challenge_waitroom",
        "chunk_air",
        "chunk_door",
        "chunk_ground",
        "climbing_pole",
        "clover",
        "coarse_water",
        "cobra",
        "coffin",
        "cog_floor",
        "construction_sign",
        "conveyorbelt_left",
        "conveyorbelt_right",
        "cooked_turkey",
        "cookfire",
        "couch",
        "crate",
        "crate_bombs",
        "crate_parachute",
        "crate_ropes",
        "crocman",
        "crossbow",
        "crown_statue",
        "crushing_elevator",
        "crushtrap",
        "crushtraplarge",
        "cursed_pot",
        "die",
        "diningtable",
        "dm_spawn_point",
        "dog_sign",
        "door",
        "door_drop_held",
        "door2",
        "door2_secret",
        "dresser",
        "drill",
        "duat_floor",
        "eggplant_altar",
        "eggplant_child",
        "eggplant_door",
        "elevator",
        "empress_grave",
        "empty",
        "empty_mech",
        "entrance",
        "entrance_shortcut",
        "excalibur_stone",
        "exit",
        "factory_generator",
        "falling_platform",
        "floor",
        "floor_hard",
        "forcefield",
        "forcefield_top",
        "fountain_drain",
        "fountain_head",
        "ghist_door2",
        "ghist_shopkeeper",
        "giant_frog",
        "giant_spider",
        "giantclam",
        "goldbars",
        "growable_climbing_pole",
        "growable_vine",
        "guts_floor",
        "haunted_corpse",
        "hermitcrab",
        "honey_downwards",
        "honey_upwards",
        "houyibow",
        "icefloor",
        "idol",
        "idol_floor",
        "idol_hold",
        "imp",
        "jiangshi",
        "jumpdog",
        "jungle_floor",
        "jungle_spear_trap",
        "key",
        "kingu",
        "ladder",
        "ladder_plat",
        "lamassu",
        "lamp_hang",
        "landmine",
        "laser_trap",
        "lava",
        "lavamander",
        "leprechaun",
        "lightarrow",
        "littorch",
        "litwalltorch",
        "locked_door",
        "lockedchest",
        "madametusk",
        "mantrap",
        "mattock",
        "merchant",
        "minewood_floor",
        "minewood_floor_hanging_hide",
        "minewood_floor_noreplace",
        "minister",
        "moai_statue",
        "mosquito",
        "mother_statue",
        "mothership_floor",
        "mummy",
        "mushroom_base",
        "necromancer",
        "nonreplaceable_babylon_floor",
        "nonreplaceable_floor",
        "octopus",
        "oldhunter",
        "olmec",
        "olmecship",
        "olmite",
        "pagoda_floor",
        "pagoda_platform",
        "palace_bookcase",
        "palace_candle",
        "palace_chandelier",
        "palace_entrance",
        "palace_floor",
        "palace_table",
        "palace_table_tray",
        "pen_floor",
        "pen_locked_door",
        "pillar",
        "pipe",
        "plasma_cannon",
        "platform",
        "pot",
        "potofgold",
        "powder_keg",
        "push_block",
        "quicksand",
        "regenerating_block",
        "robot",
        "rock",
        "royal_jelly",
        "scorpion",
        "shop_door",
        "shop_item",
        "shop_pagodawall",
        "shop_sign",
        "shop_wall",
        "shop_woodwall",
        "shopkeeper",
        "shopkeeper_vat",
        "shortcut_station_banner",
        "sidetable",
        "singlebed",
        "sister",
        "sleeping_hiredhand",
        "slidingwall_ceiling",
        "slidingwall_switch",
        "snake",
        "snap_trap",
        "sorceress",
        "spark_trap",
        "spikes",
        "spring_trap",
        "stagnant_lava",
        "starting_exit",
        "sticky_trap",
        "stone_floor",
        "storage_floor",
        "storage_guy",
        "styled_floor",
        "sunken_floor",
        "surface_floor",
        "surface_hidden_floor",
        "telescope",
        "temple_floor",
        "thief",
        "thinice",
        "thorn_vine",
        "tiamat",
        "tikiman",
        "timed_forcefield",
        "timed_powder_keg",
        "tomb_floor",
        "treasure",
        "treasure_chest",
        "treasure_vaultchest",
        "tree_base",
        "turkey",
        "tv",
        "udjat_socket",
        "ufo",
        "upsidedown_spikes",
        "ushabti",
        "vault_wall",
        "vine",
        "vlad",
        "vlad_floor",
        "walltorch",
        "wanted_poster",
        "water",
        "witchdoctor",
        "woodenlog_trap",
        "woodenlog_trap_ceiling",
        "yama",
        "yang",
        "yeti",
        "zoo_exhibit",
    ]
)

NAME_PADDING = max(map(len, VALID_TILE_CODES)) + 4
PERCENT_DELIM = re.compile(r"%\d{1,2}%?")


class TileCodes:
    def __init__(self):
        self._inner = OrderedDict()
        self.comment = None

    def all(self):
        return list(self._inner.values())

    def get(self, name):
        TileCode.validate_name(name)
        return self._inner.get(name)

    def set_obj(self, obj: "TileCode"):
        obj.validate()
        self._inner[obj.name] = obj

    def write(self, handle: TextIO):
        if self.comment:
            handle.write(f"{self.comment}\n")
        for obj in self._inner.values():
            handle.write(obj.to_line())
        handle.write("\n")


@dataclass
class TileCode:
    prefix: ClassVar[str] = DirectivePrefixes.TILE_CODE.value
    name: str
    value: str
    comment: Optional[str]

    @classmethod
    def parse(cls, line: str) -> "TileCode":
        rest, comment = split_comment(line)
        directive, value = rest.split(None, 1)
        name = directive[2:]

        if not name:
            raise ValueError("Directive missing name.")

        obj = cls(name, value, comment)
        obj.validate()

        return obj

    @staticmethod
    def validate_name(name: str):
        for part in PERCENT_DELIM.split(name):
            # names can have foo%50 where an empty rightside is valid.
            if not part:
                continue

            if part not in VALID_TILE_CODES:
                raise ValueError(f"Name {name!r} isn't a valid tile code.")

    def validate_value(self):
        if len(self.value) != 1:
            raise ValueError(
                f"Tilecode {self.name!r} has value {self.value!r} that's more than one character."
            )

    def validate(self):
        self.validate_name(self.name)
        self.validate_value()

    def to_line(self) -> str:
        return to_line(
            self.prefix, self.name, NAME_PADDING, self.value, 4, self.comment
        )

    def write(self, handle: TextIO):
        handle.write(self.to_line())



import hashlib
import logging
import os
from collections import defaultdict
from concurrent.futures import wait
from concurrent.futures.thread import ThreadPoolExecutor
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from struct import pack, unpack

import zstandard as zstd
from PIL import Image

from modlunky2.assets.constants import KNOWN_TEXTURES_V1
from modlunky2.assets.exc import NonSiblingAsset
from modlunky2.sprites.sprite_loaders import get_all_sprite_loaders
from modlunky2.sprites.sprite_mergers import get_all_sprite_mergers

from .chacha import Key, chacha, hash_filepath
from .constants import (
    BANK_ALIGNMENT,
    DDS_PNGS,
    DEFAULT_COMPRESSION_LEVEL,
    FILENAMES_TO_FILEPATHS,
    KNOWN_FILEPATHS,
    PNG_NAMES_TO_DDS_NAMES,
)
from .soundbank import extract_soundbank
from .converters import dds_to_png, png_to_dds, rgba_to_png
from .exc import FileConflict, MissingAsset, MultipleMatchingAssets
from .string_hashing import StringHashes
from .hashing import md5sum_path


logger = logging.getLogger("modlunky2")


@dataclass
class ExeAssetBlock:
    """ Represent a block of information about an asset in the exe."""

    # Position in the exe where asset is read from, or written to.
    offset: int

    # Size length of the filepath and filepath_hash
    filepath_len: int

    # The hash of the filepath
    filepath_hash: bytes

    # Whether the asset is encrypted and compressed.
    is_encrypted: bool

    # Position into the exe where the data itself
    # is read fromo, or written to.
    asset_offset: int

    # The size of the asset itself.
    asset_len: int

    @property
    def data_len(self):
        # The size of the full data portion (asset + encrypted bytes)
        return self.asset_len + 1

    @property
    def total_size(self):
        return (
            8  # 8 bytes for header (asset_len, filepath_len)
            + self.filepath_len  # The size of the filepath
            + 1  # The byte for whether the asset is encrypted
            + self.asset_len  # The size of the asset
        )

    @classmethod
    def from_exe_handle(cls, exe_handle):
        """From an exe file handle, construct an AssetInfo object.

        The exe handle should already be seeked to the appropriate position to
        begin reading.

        Returns None if there is no more assets as the current offset.
        """
        offset = exe_handle.tell()
        data_len, filepath_len = unpack(b"<II", exe_handle.read(8))

        if (data_len, filepath_len) == (0, 0):
            return

        if data_len <= 0:
            raise RuntimeError(f"Expected data length > 0, found {data_len}")

        filepath_hash = exe_handle.read(filepath_len)
        is_encrypted = exe_handle.read(1) == b"\x01"
        asset_offset = exe_handle.tell()
        asset_len = data_len - 1

        exe_handle.seek(asset_len, 1)

        return ExeAssetBlock(
            offset=offset,
            filepath_len=filepath_len,
            filepath_hash=filepath_hash,
            is_encrypted=is_encrypted,
            asset_offset=asset_offset,
            asset_len=asset_len,
        )

    def read_data(self, exe_handle):
        exe_handle.seek(self.asset_offset)
        return exe_handle.read(self.asset_len)

    def write_data(self, exe_handle, data):
        if self.asset_len != len(data):
            raise RuntimeError(
                f"Data passed was size {len(data)}, expected {self.asset_len}"
            )

        exe_handle.write(pack("<II", self.data_len, self.filepath_len))
        exe_handle.write(self.filepath_hash)
        exe_handle.write(pack("<b", self.is_encrypted))
        exe_handle.write(data)


class ExeAsset:
    def __init__(self, asset_block, filepath):
        self.asset_block = asset_block
        self.filepath = filepath
        self.data = None
        self.disk_asset = None

    def match_hash(self, hash_):
        min_len = min(len(hash_), self.asset_block.filepath_len)
        return hash_[:min_len] == self.asset_block.filepath_hash[:min_len]

    def load_data(self, handle):
        """ Cache data on the asset. Must be called before extraction."""
        handle.seek(self.asset_block.asset_offset)
        self.data = handle.read(self.asset_block.asset_len)

    def extract(
        self,
        extract_dir: Path,
        compressed_dir: Path,
        key: Key,
        compression_level=DEFAULT_COMPRESSION_LEVEL,
        recompress=True,
    ):
        if not self.filepath:
            raise RuntimeError("Asset doesn't have filepath.")

        if self.data is None:
            raise RuntimeError("load_data hasn't been called.")

        filepath = extract_dir / self.filepath

        compressed_filepath = compressed_dir / f"{self.filepath}.zst"
        md5sum_filepath = compressed_dir / f"{self.filepath}.md5sum"

        if self.asset_block.is_encrypted:
            try:
                # Decrypt
                self.data = chacha(self.filepath.encode(), self.data, key)

                # Decompress
                cctx = zstd.ZstdDecompressor()
                self.data = cctx.decompress(self.data)

                if recompress:
                    # Recompress at higher compression level to give
                    # better chance of assets fitting in binary
                    logger.info("Storing compressed asset %s...", compressed_filepath)
                    with compressed_filepath.open("wb") as compressed_file:
                        cctx = zstd.ZstdCompressor(level=compression_level)
                        compressed_data = cctx.compress(self.data)
                        compressed_file.write(compressed_data)

            except Exception:  # pylint: disable=broad-except
                logger.exception("Failed compression")
                return None

        if self.filepath in KNOWN_TEXTURES_V1:
            self.data = rgba_to_png(self.data)
        elif self.filepath in DDS_PNGS:
            self.data = dds_to_png(self.data)

        if recompress:
            # Get a hash of the the uncompressed file to be used
            # to detect if the source file changed
            md5sum = hashlib.md5(self.data).hexdigest()
            with md5sum_filepath.open("w") as md5sum_file:
                md5sum_file.write(md5sum)

        logger.info("Storing asset %s...", filepath)
        if self.filepath in DDS_PNGS:
            filepath = filepath.with_suffix(".png")

        with filepath.open("wb") as asset_file:
            asset_file.write(self.data)


class AssetStore:
    """ Represents a bundle of asset blocks read from, or to be packed into, an exe."""

    BUNDLE_OFFSET = 0x400

    def __init__(self, exe_handle):
        self.assets = []
        self.exe_handle = exe_handle
        self.total_size = 0
        self._key = Key()

    @property
    def key(self):
        return self._key.key

    def update_key(self, size):
        self._key.update(size)

    @classmethod
    def load_from_file(cls, exe_handle):
        asset_store = cls(exe_handle)
        asset_store.exe_handle.seek(cls.BUNDLE_OFFSET)

        while True:
            asset_block = ExeAssetBlock.from_exe_handle(exe_handle)

            if asset_block is None:
                # We've reached the end of the asset blocks.
                break

            asset_store.update_key(asset_block.data_len)
            asset_store.total_size += asset_block.total_size
            asset_store.assets.append(ExeAsset(asset_block, None))

        asset_store.populate_asset_filepaths()
        return asset_store

    def find_asset(self, filepath):
        if filepath is None:
            return None
        filepath_hash = self.hash_filepath(filepath)
        for asset in self.assets:
            if asset.match_hash(filepath_hash):
                return asset
        return None

    def hash_filepath(self, filepath):
        if filepath is None:
            return None

        if not isinstance(filepath, bytes):
            filepath = filepath.encode()

        return hash_filepath(filepath, self.key)

    def populate_asset_filepaths(self):
        for filepath in KNOWN_FILEPATHS:
            asset = self.find_asset(filepath)
            if asset is None:
                continue
            asset.filepath = filepath

    @staticmethod
    def _extract_single(asset, *args, **kwargs):
        try:
            logger.info("Extracting %s... ", asset.filepath)
            asset.extract(*args, **kwargs)
        except Exception:  # pylint: disable=broad-except
            logger.exception("Failed Extraction")

    @staticmethod
    def _merge_single_entity(sprite_merger, sprite_loaders):
        sprite_merger.do_merge(sprite_loaders)
        sprite_merger.save()

    def extract(
        self,
        extract_dir,
        compressed_dir,
        compression_level=DEFAULT_COMPRESSION_LEVEL,
        max_workers=max(os.cpu_count() - 2, 1),
        recompress=True,
        generate_string_hashes=True,
        create_entity_sheets=True,
        extract_sound_extensions=None,
        reuse_extracted=False,
    ):
        unextracted = []

        if not reuse_extracted:
            for asset in self.assets:
                if asset.filepath is None:
                    # No known filepaths matched this asset.
                    unextracted.append(asset)
                    continue

                asset.load_data(self.exe_handle)

            with ThreadPoolExecutor(max_workers=max_workers) as pool:
                futures = [
                    pool.submit(
                        self._extract_single,
                        asset,
                        extract_dir,
                        compressed_dir,
                        self.key,
                        compression_level,
                        recompress,
                    )
                    for asset in self.assets
                    if asset.filepath
                ]
                wait(futures, timeout=300)

        if generate_string_hashes:
            self.hash_strings(extract_dir)

        if create_entity_sheets:
            logger.info("Creating entity sprite sheets...")
            sprite_loaders = get_all_sprite_loaders(extract_dir)
            sprite_mergers = get_all_sprite_mergers(extract_dir)

            with ThreadPoolExecutor(max_workers=max_workers) as pool:
                futures = [
                    pool.submit(
                        self._merge_single_entity,
                        sprite_merger,
                        sprite_loaders,
                    )
                    for sprite_merger in sprite_mergers
                ]
                wait(futures, timeout=300)

            logger.info("Done creating entity sprite sheets...")

        if extract_sound_extensions:
            extract_soundbank(
                extract_dir / "soundbank.bank",
                extract_dir / "soundbank",
                extract_sound_extensions,
            )

        return unextracted

    def hash_strings(self, extract_dir):
        # Use the english strings as the source for generating hash values.
        english_strings = self.find_asset("strings00.str")
        if english_strings is None or english_strings.data is None:
            logging.warning("Didn't find data for english strings in strings00.str")
            return

        string_hashes = StringHashes.from_data(english_strings.data)

        # Create the hashed string files separately since they all depend on strings00.str
        for asset in self.assets:
            if asset.filepath and asset.filepath.endswith(".str"):
                asset_file_path = Path(asset.filepath)
                hashed_strings_file = (
                    extract_dir
                    / asset_file_path.parent
                    / f"{asset_file_path.stem}_hashed{asset_file_path.suffix}"
                )
                string_hashes.write_string_hashes(asset.data, hashed_strings_file)

    def pack_assets(self):
        self.exe_handle.seek(self.BUNDLE_OFFSET)

        for asset in self.assets:
            if asset.filepath is None:
                continue

            assert asset.asset_block.asset_len == asset.disk_asset.get_asset_len()
            data = asset.disk_asset.get_asset_data()

            if asset.asset_block.is_encrypted:
                logger.info("Encrypting file %s", asset.disk_asset.asset_path)
                data = chacha(asset.filepath.encode(), data, self.key)

            logger.info("Packing file %s", asset.disk_asset.asset_path)
            self.exe_handle.write(
                pack("<II", asset.asset_block.data_len, asset.asset_block.filepath_len)
            )
            self.exe_handle.write(asset.asset_block.filepath_hash)
            self.exe_handle.write(pack("<b", asset.asset_block.is_encrypted))
            self.exe_handle.write(data)

        self.exe_handle.write(pack("<II", 0, 0))

    def recalculate_key(self):
        """ Recalculate the key from the current assets."""
        new_key = Key()
        for asset in self.assets:
            if asset.filepath is None:
                continue
            new_key.update(asset.asset_block.data_len)
        self._key = new_key

    def update_filepath_hashes(self):
        for asset in self.assets:
            if asset.filepath is None:
                continue
            asset.asset_block.filepath_hash = self.hash_filepath(asset.filepath).ljust(
                asset.asset_block.filepath_len, b"\x00"
            )

    def repackage(
        self,
        search_dirs,
        fallback_dir,
        compressed_dir,
        compression_level=DEFAULT_COMPRESSION_LEVEL,
    ):
        disk_bundle = DiskBundle.from_dirs(
            self.assets,
            search_dirs,
            fallback_dir,
            compressed_dir,
        )
        disk_bundle.compress_if_needed(compression_level=compression_level)

        offset = self.BUNDLE_OFFSET
        for asset in self.assets:
            if asset.filepath is None:
                continue
            disk_asset = disk_bundle.get(str(Path(asset.filepath).name))
            if disk_asset is None:
                raise MissingAsset(f"FAIL {asset.filepath}")

            asset.disk_asset = disk_asset

            asset.asset_block.offset = offset
            asset.asset_block.asset_len = disk_asset.get_asset_len()
            asset.asset_block.asset_offset = (
                asset.asset_block.offset + 8 + asset.asset_block.filepath_len + 1
            )

            # The name hash of soundbank files is padded such that the asset_offset
            # is divisible by 32.
            #
            # Padding is between 1 and 32 bytes
            if disk_asset.asset_path.suffix == ".bank":
                padding = (
                    BANK_ALIGNMENT - asset.asset_block.asset_offset % BANK_ALIGNMENT
                )
                asset.asset_block.filepath_len += padding
                asset.asset_block.asset_offset += padding

            offset += asset.asset_block.total_size

        self.recalculate_key()
        self.update_filepath_hashes()
        self.pack_assets()


class ResolutionPolicy(Enum):
    RaiseError = 1
    FirstWins = 2
    LastWins = 3


@dataclass
class DiskBundle:
    def __init__(self, disk_assets):
        self.disk_assets = disk_assets

    def get(self, filepath, default=None):
        return self.disk_assets.get(filepath, default)

    @staticmethod
    def get_files_from_search_dir(search_dir: Path):
        out_files = {}

        if not search_dir.exists():
            return out_files

        for root, dirs, files in os.walk(search_dir, topdown=True):
            # Remove compressed directories
            dirs[:] = [d for d in dirs if d not in [".compressed"]]

            for file_ in files:
                # For DDS_PNGS we need to convert to actual name to test
                # for validity of files
                real_name = PNG_NAMES_TO_DDS_NAMES.get(file_, file_)
                if real_name not in FILENAMES_TO_FILEPATHS:
                    continue

                if file_ in out_files:
                    raise MultipleMatchingAssets(
                        f"Found {file_} multiple times in {search_dir}"
                    )

                out_files[file_] = Path(root) / file_

        return out_files

    @classmethod
    def from_dirs(
        cls,
        exe_assets,
        search_dirs,
        fallback_dir,
        compressed_dir,
        resolution_policy=ResolutionPolicy.RaiseError,
    ):

        modfiles_by_filename = defaultdict(list)
        disk_assets = {}

        for search_dir in search_dirs:
            for file_, file_path in cls.get_files_from_search_dir(
                Path(search_dir)
            ).items():
                modfiles_by_filename[file_].append(file_path)

        for asset in exe_assets:
            if asset.filepath is None:
                continue
            filepath = Path(asset.filepath)
            if asset.filepath in DDS_PNGS:
                filepath = filepath.with_suffix(".png")
            modpack_files = modfiles_by_filename.get(filepath.name)

            # No modpacks overrode this asset. Get it from
            # fallback directory.
            if not modpack_files:
                asset_path = fallback_dir / filepath
                if not asset_path.exists():
                    raise MissingAsset(f"Didn't find an asset for {filepath}")

                disk_assets[str(Path(asset.filepath).name)] = DiskAsset(
                    asset_path, compressed_dir, asset
                )
                continue

            if (
                resolution_policy == ResolutionPolicy.RaiseError
                and len(modpack_files) >= 2
            ):
                raise FileConflict(
                    f"{filepath} found in multiple packs: {', '.join(map(str, modpack_files))}"
                )

            idx = 0
            if resolution_policy == ResolutionPolicy.FirstWins:
                idx = 0
            elif resolution_policy == ResolutionPolicy.LastWins:
                idx = -1

            asset_path = modpack_files[idx]
            disk_assets[str(Path(asset.filepath).name)] = DiskAsset(
                asset_path, compressed_dir, asset
            )

        return cls(disk_assets)

    def compress_if_needed(
        self,
        compression_level=DEFAULT_COMPRESSION_LEVEL,
        max_workers=max(os.cpu_count() - 2, 1),
    ):
        with ThreadPoolExecutor(max_workers=max_workers) as pool:
            futures = [
                pool.submit(disk_asset.compress, compression_level)
                for disk_asset in self.disk_assets.values()
                if disk_asset.needs_compression()
            ]
            wait(futures, timeout=300)


@dataclass
class DiskAsset:
    # Path to the asset on disk
    asset_path: Path

    # Root directory where compression files are stored
    compressed_dir: Path

    exe_asset: ExeAsset

    def __post_init__(self):
        # assets cannot live higher than the compressed dir
        # to avoid complicating the caching structure. This
        # code ensures the asset_path is a sibling or a child
        # of a sibling. This also normalized the path, removing
        # and `..`'s.
        root = self.compressed_dir.resolve().parent
        try:
            self.rel_asset_path = self.asset_path.resolve().relative_to(root)
            # This is needed to clean up and `.`/`..`'s
            self.asset_path = root / self.rel_asset_path
        except ValueError as err:
            raise NonSiblingAsset(
                f"Disk Asset ({self.asset_path}) not sibling of ({self.compressed_dir})."
            ) from err

    def md5sum_of_asset(self):
        return md5sum_path(self.asset_path)

    @property
    def real_suffix(self):
        if self.exe_asset.filepath in DDS_PNGS:
            return ".DDS"
        return Path(self.exe_asset.filepath).suffix

    @property
    def compressed_path(self):
        return (
            self.compressed_dir
            / f"{self.rel_asset_path.with_suffix(self.real_suffix)}.zst"
        )

    @property
    def md5sum_path(self):
        return (
            self.compressed_dir
            / f"{self.rel_asset_path.with_suffix(self.real_suffix)}.md5sum"
        )

    def needs_compression(self):
        if not self.exe_asset.asset_block.is_encrypted:
            return False

        if not self.md5sum_path.exists():
            return True

        if not self.compressed_path.exists():
            return True

        md5sum = self.md5sum_of_asset()
        with self.md5sum_path.open("rb") as md5sum_file:
            stored_md5sum = md5sum_file.read().strip()
        if md5sum != stored_md5sum:
            return True

        return False

    def compress(self, compression_level=DEFAULT_COMPRESSION_LEVEL):
        if not self.exe_asset.asset_block.is_encrypted:
            return

        self.compressed_path.parent.mkdir(parents=True, exist_ok=True)
        self.md5sum_path.parent.mkdir(parents=True, exist_ok=True)

        if self.asset_path.suffix == ".png":
            logger.info('Converting image "%s" to DDS', self.asset_path)
            with Image.open(self.asset_path) as img:
                data = png_to_dds(img)
        else:
            with open(self.asset_path, "rb") as asset_file:
                data = asset_file.read()

        md5sum = self.md5sum_of_asset()
        with self.md5sum_path.open("wb") as md5sum_file:
            md5sum_file.write(md5sum)

        logger.info("Compressing %s...", self.asset_path)
        cctx = zstd.ZstdCompressor(level=compression_level)
        data = cctx.compress(data)
        with open(self.compressed_path, "wb") as compressed_file:
            compressed_file.write(data)

    def get_asset_len(self):
        if self.exe_asset.asset_block.is_encrypted:
            path = self.compressed_path
        else:
            path = self.asset_path
        return path.stat().st_size

    def get_asset_data(self):
        if self.exe_asset.asset_block.is_encrypted:
            path = self.compressed_path
        else:
            path = self.asset_path

        with path.open("rb") as file_:
            return file_.read()



from abc import abstractmethod
from pathlib import Path

from PIL import Image

from .base_sprite_loader import BaseSpriteLoader


class AbstractDecoSheet(BaseSpriteLoader):
    _sprite_sheet: Image.Image
    _base_path: Path

    @property
    @abstractmethod
    def biome_name(self) -> str:
        raise NotImplementedError

    @property
    def _sprite_sheet_path(self):
        return Path(f"Data/Textures/deco_{self.biome_name}.png")



import requests
from packaging import version

from modlunky2.constants import BASE_DIR


def latest_version():
    try:
        return version.parse(
            requests.get(
                "https://api.github.com/repos/spelunky-fyi/modlunky2/releases/latest"
            ).json()["tag_name"]
        )
    except Exception:  # pylint: disable=broad-except
        return None


def current_version():
    with (BASE_DIR / "VERSION").open() as version_file:
        return version.parse(version_file.read().strip())



from .big import (
    Big1,
    Big2,
    Big3,
    Big4,
    Big5,
    Big6,
    OsirisAndAlienQueen,
    Yama,
    OlmecAndMech,
)
from .ghost import Ghost
from .basic import Basic1, Basic2, Basic3, Monsters01, Monsters02, Monsters03
from .mounts import Mounts
from .pets import Pets

__all__ = [
    "Big1",
    "Big2",
    "Big3",
    "Big4",
    "Big5",
    "Big6",
    "OsirisAndAlienQueen",
    "Yama",
    "OlmecAndMech",
    "Ghost",
    "Basic1",
    "Basic2",
    "Basic3",
    "Monsters01",
    "Monsters02",
    "Monsters03",
    "Mounts",
    "Pets",
]



from .base_classes.base_deco_sheet import AbstractDecoSheet

__all__ = [
    "CaveDecoSheet",
    "VolcanaDecoSheet",
    "JungleDecoSheet",
    "TidePoolDecoSheet",
    "TempleDecoSheet",
    "IceCavesDecoSheet",
    "BabylonDecoSheet",
    "SunkenCityDecoSheet",
    "EggplantDecoSheet",
    "SurfaceDecoSheet",
]


class CaveDecoSheet(AbstractDecoSheet):
    biome_name = "cave"
    _chunk_size = 128
    _chunk_map = {"kali_bg": (0, 0, 4, 5), "woodenlog_trap": (1, 5, 3, 10)}


class VolcanaDecoSheet(AbstractDecoSheet):
    biome_name = "volcano"
    _chunk_size = 128
    _chunk_map = {
        "kali_bg": (0, 0, 4, 5),
        "drill": (1, 5, 3, 8),
        "vlad_banner": (1, 8, 3, 12),
    }


class JungleDecoSheet(AbstractDecoSheet):
    biome_name = "jungle"
    _chunk_size = 128
    _chunk_map = {"stone_door": (0, 5, 3, 7)}


class TidePoolDecoSheet(AbstractDecoSheet):
    biome_name = "tidepool"
    _chunk_size = 128
    _chunk_map = {}


class TempleDecoSheet(AbstractDecoSheet):
    biome_name = "temple"
    _chunk_size = 128
    _chunk_map = {}


class IceCavesDecoSheet(AbstractDecoSheet):
    biome_name = "ice"
    _chunk_size = 128
    _chunk_map = {
        "moai_statue": (0, 7, 4, 12),
    }


class BabylonDecoSheet(AbstractDecoSheet):
    biome_name = "babylon"
    _chunk_size = 128
    _chunk_map = {}


class SunkenCityDecoSheet(AbstractDecoSheet):
    biome_name = "sunken"
    _chunk_size = 128
    _chunk_map = {
        "mother_statue": (0, 5, 2, 11),
    }


class EggplantDecoSheet(AbstractDecoSheet):
    biome_name = "eggplant"
    _chunk_size = 128
    _chunk_map = {}


class SurfaceDecoSheet(AbstractDecoSheet):
    biome_name = "basecamp"
    _chunk_size = 128
    _chunk_map = {
        "construction_sign": (5, 12, 6, 13),
        "singlebed": (0, 15, 1, 16),
        "dresser": (2, 14, 3, 15),
        "bunkbed": (0, 13, 2, 15),
        "diningtable": (14, 10, 16, 11),
        "sidetable": (9, 12, 10, 13),
        "chair_looking_left": (3, 15, 4, 16),
        "chair_looking_right": (3, 15, 4, 16),
        "couch": (6, 15, 8, 16),
        "tv": (8, 15, 9, 16),
        "dog_sign": (12, 10, 14, 12),
        "shortcut_station_banner": (10, 8, 12, 12),
        "telescope": (12, 8, 14, 10),
    }



from chunk import Chunk


class RIFFChunk(Chunk):
    def __init__(self, f):
        super().__init__(f, align=True, bigendian=False)
        self.skip()
        self.end = self.tell() + self.offset
        self.seek(0)
        self.children = None
        self.type = None

    def populate_children(self):
        children = []
        while self.file.tell() < self.end:
            chunk = RIFFChunk(self.file)
            if chunk.getname() == b"LIST":
                chunk.type = chunk.read(4)
                chunk.populate_children()
            chunk.seek(0)
            chunk.skip()
            children.append(chunk)
        self.children = children

    def __repr__(self):
        return "<%s %r with %s children>" % (
            self.__class__.__name__,
            self.getname(),
            "no" if self.children is None else len(self.children),
        )


class RIFF(RIFFChunk):
    def __init__(self, f):
        super().__init__(f)

        assert self.getname() == b"RIFF"
        self.seek(0)
        self.type = self.read(4)
        self.populate_children()

    def save(self):
        pass



from pathlib import Path

from ..base_classes.base_sprite_loader import BaseSpriteLoader
from ..util import chunks_from_animation


class Pets(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/monsters_pets.png")
    _chunk_size = 128
    _chunk_map = {
        "monty": (0, 0, 1, 1),
        **chunks_from_animation("monty_row_0_", (0, 0, 1, 1), 12),
        **chunks_from_animation("monty_row_1_", (0, 1, 1, 2), 10),
        **chunks_from_animation("monty_row_2_", (0, 2, 1, 3), 12),
        **chunks_from_animation("monty_row_3_", (0, 3, 1, 4), 12),
        "percy": (0, 4, 1, 5),
        **chunks_from_animation("percy_row_0_", (0, 4, 1, 5), 12),
        **chunks_from_animation("percy_row_1_", (0, 5, 1, 6), 10),
        **chunks_from_animation("percy_row_2_", (0, 6, 1, 7), 12),
        **chunks_from_animation("percy_row_3_", (0, 7, 1, 8), 12),
        "poochi": (0, 8, 1, 9),
        **chunks_from_animation("poochi_row_0_", (0, 8, 1, 9), 12),
        **chunks_from_animation("poochi_row_1_", (0, 9, 1, 10), 10),
        **chunks_from_animation("poochi_row_2_", (0, 10, 1, 11), 12),
        **chunks_from_animation("poochi_row_3_", (0, 11, 1, 12), 12),
    }



from pathlib import Path

from .base_classes import BaseSpriteLoader


class CoffinSheet(BaseSpriteLoader):
    _sprite_sheet_path = Path("Data/Textures/coffins.png")
    _chunk_size = 128
    _chunk_map = {
        "coffin": (0, 0, 2, 2),
    }



import re
from dataclasses import dataclass
from enum import Enum
from collections import OrderedDict
from itertools import zip_longest
from typing import ClassVar, List, Optional, TextIO, Tuple

from modlunky2.levels.utils import split_comment

from .utils import DirectivePrefixes

VALID_LEVEL_TEMPLATES = set(
    [
        "abzu_backdoor",
        "altar",
        "anubis_room",
        "apep",
        "beehive",
        "beehive_entrance",
        "blackmarket_coffin",
        "blackmarket_entrance",
        "blackmarket_exit",
        "cache",
        "cavemanshop",
        "challenge_0-0",
        "challenge_0-1",
        "challenge_0-2",
        "challenge_0-3",
        "challenge_1-0",
        "challenge_1-1",
        "challenge_1-2",
        "challenge_1-3",
        "challenge_bottom",
        "challenge_entrance",
        "challenge_special",
        "chunk_air",
        "chunk_door",
        "chunk_ground",
        "coffin_player",
        "coffin_player_vertical",
        "coffin_unlockable",
        "cog_altar_top",
        "crashedship_entrance",
        "crashedship_entrance_notop",
        "curioshop",
        "diceshop",
        "empress_grave",
        "entrance",
        "entrance_drop",
        "exit",
        "exit_notop",
        "feeling_factory",
        "feeling_prison",
        "feeling_tomb",
        "ghistroom",
        "ghistshop",
        "idol",
        "idol_top",
        "lake_exit",
        "lake_normal",
        "lake_notop",
        "lakeoffire_back_entrance",
        "lakeoffire_back_exit",
        "machine_bigroom_path",
        "machine_bigroom_side",
        "machine_keyroom",
        "machine_rewardroom",
        "machine_tallroom_path",
        "machine_tallroom_side",
        "machine_wideroom_path",
        "machine_wideroom_side",
        "moai",
        "mothership_coffin",
        "mothership_entrance",
        "mothership_exit",
        "motherstatue_room",
        "oldhunter_cursedroom",
        "oldhunter_keyroom",
        "oldhunter_rewardroom",
        "olmecship_room",
        "palaceofpleasure_0-0",
        "palaceofpleasure_0-1",
        "palaceofpleasure_0-2",
        "palaceofpleasure_1-0",
        "palaceofpleasure_1-1",
        "palaceofpleasure_1-2",
        "palaceofpleasure_2-0",
        "palaceofpleasure_2-1",
        "palaceofpleasure_2-2",
        "palaceofpleasure_3-0",
        "palaceofpleasure_3-1",
        "palaceofpleasure_3-2",
        "passage_horz",
        "passage_turn",
        "passage_vert",
        "path_drop",
        "path_drop_notop",
        "path_normal",
        "path_notop",
        "pen_room",
        "posse",
        "quest_thief1",
        "quest_thief2",
        "room2",
        "shop",
        "shop_attic",
        "shop_basement",
        "shop_entrance_down",
        "shop_entrance_up",
        "side",
        "sisters_room",
        "storage_room",
        "tuskdiceshop",
        "tuskfrontdiceshop",
        "udjatentrance",
        "udjattop",
        "ushabti_entrance",
        "ushabti_room",
        "vault",
        "vlad_bottom_exit",
        "vlad_bottom_tunnel",
        "vlad_drill",
        "vlad_entrance",
        "vlad_tunnel",
    ]
)

SETROOM_RE = re.compile(r"^setroom\d{1,2}-\d{1,2}")

TEMPLATE_COMMENT = "/" * 80


class TemplateSetting(Enum):
    IGNORE = "ignore"
    FLIP = "flip"
    ONLYFLIP = "onlyflip"
    DUAL = "dual"
    RARE = "rare"
    HARD = "hard"
    LIQUID = "liquid"
    PURGE = "purge"

    def to_line(self):
        prefix = DirectivePrefixes.TEMPLATE_SETTING.value
        return f"{prefix}{self.value}\n"


VALID_TEMPLATE_SETTINGS = set(rf"\!{setting.value}" for setting in TemplateSetting)


class LevelTemplates:
    def __init__(self):
        self._inner = OrderedDict()
        self.comment = None

    def all(self):
        return list(self._inner.values())

    def get(self, name):
        LevelTemplate.validate_name(name)
        return self._inner.get(name)

    def set_obj(self, obj: "LevelTemplate"):
        obj.validate()
        self._inner[obj.name] = obj

    def write(self, handle: TextIO):
        if self.comment:
            handle.write(f"{self.comment}\n")
        for idx, template in enumerate(self._inner.values()):
            template.write(handle)
            if idx < len(self._inner) - 1:
                handle.write("\n")


@dataclass
class Chunk:
    comment: Optional[str]
    settings: List[TemplateSetting]
    foreground: List[List[str]]
    background: List[List[str]]

    @staticmethod
    def partition_line(line: str) -> Tuple[str, str]:
        foreground, _, background = line.partition(" ")
        return foreground.strip(), background.strip()

    @classmethod
    def parse(cls, file_handle: TextIO) -> "Chunk":
        chunk = cls(comment="", settings=[], foreground=[], background=[])

        for line in file_handle:
            line = line.strip()

            if not line:
                # Reached the end of the file chunk or file.
                return chunk

            if line.startswith("//"):
                chunk.comment += f"{line}"
            elif line in VALID_TEMPLATE_SETTINGS:
                chunk.settings.append(TemplateSetting(line[2:]))
            else:
                foreground, background = cls.partition_line(line)
                if background:
                    chunk.background.append(list(background))
                chunk.foreground.append(list(foreground))

        return chunk

    def write(self, handle: TextIO):
        if self.comment:
            handle.write(f"// {self.comment.rstrip()}\n")

        for setting in self.settings:
            handle.write(setting.to_line())

        for fg_line, bg_line in zip_longest(self.foreground, self.background):
            line = "".join(fg_line)
            if bg_line:
                line = f"{line} {''.join(bg_line)}"
            handle.write(f"{line}\n")


@dataclass
class LevelTemplate:
    prefix: ClassVar[str] = DirectivePrefixes.TEMPLATE.value

    name: str
    comment: Optional[str]
    chunks: List[Chunk]

    @classmethod
    def parse(cls, line: str, file_handle: TextIO) -> "LevelTemplate":
        directive, comment = split_comment(line)
        name = directive[2:]

        if not name:
            raise ValueError("Directive missing name.")

        chunks = []
        level_template = cls(name, comment, chunks)

        next_line = file_handle.peek()
        while next_line:
            next_line, comment = split_comment(next_line)

            # We've reached the next Template, return
            if next_line.startswith(DirectivePrefixes.TEMPLATE.value):
                return level_template

            if next_line or comment:
                chunk = Chunk.parse(file_handle)
                chunks.append(chunk)
            else:
                # Advance the cursor and peek the next line.
                file_handle.advance()

            next_line = file_handle.peek()

        return level_template

    @staticmethod
    def validate_name(name: str):
        if name not in VALID_LEVEL_TEMPLATES and not SETROOM_RE.match(name):
            raise ValueError(f"Name {name!r} isn't a valid level template")

    def validate(self):
        self.validate_name(self.name)

    def write(self, handle: TextIO):
        handle.write(f"{TEMPLATE_COMMENT}\n")
        name_line = f"{self.prefix}{self.name}"
        if self.comment:
            name_line += f"   // {self.comment}"
        handle.write(f"{name_line}\n")
        handle.write(f"{TEMPLATE_COMMENT}\n\n")

        for idx, chunk in enumerate(self.chunks):
            chunk.write(handle)
            if idx < len(self.chunks) - 1:
                handle.write("\n")



from abc import ABC, abstractmethod
from functools import wraps
from pathlib import Path
from typing import Callable, Dict, Optional, Union

from PIL import Image

from .types import chunk_map_type

_DEFAULT_BASE_PATH = Path(
    r"C:\Program Files (x86)\Steam\steamapps\common\Spelunky 2\Mods\Extracted"
)

_CACHED_NONE_SENTINEL = object()


def _cache_img_not_class(f):
    """using this because functools.lrucache will keep the class object from getting
    GCed if you are using it on a method"""

    # noinspection PyProtectedMember
    @wraps(f)
    def cache_img(slf, name: str):
        img = slf._cache_dict.get(name)
        if not img:
            img = f(slf, name) or _CACHED_NONE_SENTINEL
            slf._cache_dict[name] = img
        if img is _CACHED_NONE_SENTINEL:
            return None
        return img

    return cache_img


class BaseSpriteLoader(ABC):
    @property
    @abstractmethod
    def _sprite_sheet_path(self) -> Path:
        """
        Define the path from a "base path" to the specific png file this is for reading,
        for example for `items.png` it should return Path('Data/Textures/items.png')
        """
        pass

    @property
    @abstractmethod
    def _chunk_size(self) -> int:
        """
        define how many pixels wide/tall each piece is expected to be
        """
        pass

    @property
    @abstractmethod
    def _chunk_map(self) -> chunk_map_type:
        """
        Define the mapping of name, coordinates in a dictionary in the form of:
        `"thing_name": (1, 0, 2, 2)`
        The digits will be multiplied by the _chunk_size when the piece is accessed.
        """
        pass

    def __init__(self, base_path: Path = _DEFAULT_BASE_PATH):
        self.base_path = base_path
        self._sprite_sheet = Image.open(self.base_path / self._sprite_sheet_path)
        # This is to support the caching decorator and let it use storage on this class
        # otherwise we get doofy global dicts for caching
        self._cache_dict = {}

    def _get_block(
        self,
        left: Union[int, float],
        upper: Union[int, float],
        right: Union[int, float],
        lower: Union[int, float],
    ) -> Image.Image:
        """Used to get chunks of the sprite sheet."""
        bbox = tuple(map(lambda x: x * self._chunk_size, (left, upper, right, lower)))
        return self._sprite_sheet.crop(bbox)

    @_cache_img_not_class
    def get(self, name: str) -> Optional[Image.Image]:
        coords = self._chunk_map.get(name)
        if coords:
            return self._get_block(*coords)

    def key_map(self) -> Dict[str, Callable]:
        return {k: self.get for k in self._chunk_map}



import argparse
import logging
from pathlib import Path

from appdirs import user_config_dir, user_data_dir

from .ui import ModlunkyUI
from .config import Config

APP_AUTHOR = "spelunky.fyi"
APP_NAME = "modlunky2"

logger = logging.getLogger("modlunky2")


def make_dirs(*dirs):
    for dir_ in dirs:
        if not dir_.exists():
            dir_.mkdir(parents=True, exist_ok=True)


def main():
    parser = argparse.ArgumentParser(description="Tool for modding Spelunky 2.")
    parser.add_argument(
        "--install-dir",
        default=None,
        help="Path to Spelunky 2 installation. (Default: %(default)s",
    )
    parser.add_argument(
        "--beta", default=False, action="store_true", help="Display beta features."
    )
    args = parser.parse_args()

    log_format = "%(asctime)s: %(message)s"
    logging.basicConfig(format=log_format, level=logging.INFO, datefmt="%H:%M:%S")
    launch(args)


def launch(args):
    config_dir = Path(user_config_dir(APP_NAME, APP_AUTHOR))
    data_dir = Path(user_data_dir(APP_NAME, APP_AUTHOR))

    make_dirs(config_dir, data_dir)
    config = Config.from_path(config_dir / "config.json")

    if args.install_dir:
        config.install_dir = Path(args.install_dir)

    if args.beta:
        config.beta = True

    native_ui = ModlunkyUI(config, data_dir)
    native_ui.mainloop()
